from silence.settings import settings

###############################################################################
# Utility class for printing a pretty API endpoint tree
###############################################################################
class APITree:
    def __init__(self):
        self.tree = {}

    def add_url(self, url):
        root = settings.API_PREFIX
        # Strip the API prefix and split by /
        url = url[len(root):]
        if url.startswith("/"):
            url = url[1:]
        spl = url.split("/")
        spl = ["/" + x for x in spl]

        # Update the internal tree info
        if root not in self.tree:
            self.tree[root] = []
        if spl[0] not in self.tree[root]:
            self.tree[root].append(spl[0])

        for token, next_token in zip(spl, spl[1:]):
            if token not in self.tree:
                self.tree[token] = [next_token]
            elif next_token not in self.tree[token]:
                self.tree[token].append(next_token)

    # From https://stackoverflow.com/questions/51903172/how-to-display-a-tree-in-python-similar-to-msdos-tree-command
    def format_tree(self, indent_width=4):
        def _ptree(res, start, parent, tree, grandpa=None, indent=""):
            if parent != start:
                if grandpa is None:  # Ask grandpa kids!
                    res.add(str(parent))
                else:
                    res.add(str(parent) + "\n")
            if parent not in tree:
                return
            for child in tree[parent][:-1]:
                res.add(indent + "├" + "─" * indent_width)
                _ptree(res, start, child, tree, parent, indent + "│" + " " * 4)
            child = tree[parent][-1]
            res.add(indent + "└" + "─" * indent_width)
            _ptree(res, start, child, tree, parent, indent + " " * 5)  # 4 -> 5

        res = _StrCont("")
        start = settings.API_PREFIX
        parent = start
        _ptree(res, start, parent, self.tree)
        return res.get()

class _StrCont:
    def __init__(self, string):
        self.string = string

    def add(self, string):
        self.string += string

    def get(self):
        return self.string


from random import choice
import pymysql

from silence.settings import settings

###############################################################################
# The connector fetches the relevant configuration parameters
# and uses them to build a connection to the database.
###############################################################################

global conn_pool
conn_pool = None

def get_conn():
    # Create the pool of connections lazily
    global conn_pool
    if conn_pool is None:
        conn_pool = [pymysql.connect(
                        host=settings.DB_CONN["host"],
                        port=settings.DB_CONN["port"],
                        user=settings.DB_CONN["username"],
                        password=settings.DB_CONN["password"],
                        database=settings.DB_CONN["database"],
                    ) for _ in range(settings.DB_CONN_POOL_SIZE)]
    return choice(conn_pool)



import logging
import traceback

import colorama
from colorama import Fore, Style

colorama.init()

class DefaultFormatter(logging.Formatter):

    format_nolvl = "%(message)s"
    format_lvl = "[%(levelname)s] %(message)s"

    FORMATS = {
        logging.DEBUG: Style.BRIGHT + Fore.BLACK + format_lvl + Style.RESET_ALL,
        logging.INFO: format_nolvl,
        logging.WARNING: Style.BRIGHT + Fore.YELLOW + format_lvl + Style.RESET_ALL,
        logging.ERROR: Style.BRIGHT + Fore.RED + format_lvl + Style.RESET_ALL,
        logging.CRITICAL: Style.BRIGHT + Fore.RED + format_lvl + Style.RESET_ALL
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        res = formatter.format(record)
        
        # If we're logging an exception, print the stack trace in red
        if record.exc_info:
            res = res.replace(Style.RESET_ALL, "")
            res += Style.RESET_ALL

        return res

    



import argparse
import sys

from silence.cli.commands import run, createdb, new

###############################################################################
# Command line interface manager
# It provides the entry point for the 'silence' command.
###############################################################################

HANDLERS = {
    "run": run.handle,
    "createdb": createdb.handle,
    "new": new.handle
}

def run_from_command_line():
    parser = argparse.ArgumentParser(
        description="Silence: An educational framework for deploying RESTful APIs and Web applications."
    )
    
    # Force the user to select one of the available commands,
    # and allow them to provide additional options after it.
    parser.add_argument("command", choices=["run", "createdb", "new"])
    parser.add_argument("options", nargs="*")

    # Show the help dialog if the command is issued without any arguments
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args()
    HANDLERS[args.command.lower()](args.options)



from flask import jsonify, request
from werkzeug.security import check_password_hash, generate_password_hash

from silence.auth.tokens import create_token, check_token
from silence.db import dal
from silence.sql.builder import get_login_query, get_register_user_query
from silence.settings import settings
from silence.exceptions import HTTPError
from silence.logging.default_logger import logger

###############################################################################
# Defines the default endpoints provided by Silence,
# mainly /login and /register
###############################################################################

USERS_TABLE = settings.USER_AUTH_DATA["table"]
IDENTIFIER_FIELD = settings.USER_AUTH_DATA["identifier"]
PASSWORD_FIELD = settings.USER_AUTH_DATA["password"]

def login():
    # Ensure that the user has sent the required fields
    form = request.json if request.is_json else request.form
    username = form.get(IDENTIFIER_FIELD, None)
    password = form.get(PASSWORD_FIELD, None)

    if not username or not password:
        raise HTTPError(400, f"Both '{IDENTIFIER_FIELD}' and '{PASSWORD_FIELD}' are required")

    logger.debug(f"Login request from user {username} with password {password}")

    # Look if there is an user with such username
    q = get_login_query(USERS_TABLE, IDENTIFIER_FIELD, username)
    users = dal.api_safe_query(q)

    if not users:
        logger.debug(f"The identifier {username} was not found")
        raise HTTPError(400, "User not found")

    # The identifier field should be unique (/register also takes care of that)
    # so we can just extract the first one
    user = users[0]
    
    # Check if the user's password matches the provided one
    if PASSWORD_FIELD not in user:
        raise HTTPError(500, f"The user has no attribute '{PASSWORD_FIELD}'")

    password_ok = check_password_hash(user[PASSWORD_FIELD], password)
    if not password_ok:
        logger.debug(f"Incorrect password")
        raise HTTPError(400, "The password is not correct")

    # If we've reached here the login is successful, generate a session token
    # and return it with the logged user's info
    logger.debug("Login OK")
    token = create_token(user)
    del user[PASSWORD_FIELD]
    res = {"sessionToken": token, "user": user}
    return jsonify(res), 200

def register():
    # Ensure that the user has sent the required fields
    form = request.json if request.is_json else request.form
    username = form.get(IDENTIFIER_FIELD, None)
    password = form.get(PASSWORD_FIELD, None)

    if not username or not password:
        raise HTTPError(400, f"Both '{IDENTIFIER_FIELD}' and '{PASSWORD_FIELD}' are required")

    logger.debug(f"Register request with data {form}")

    # Ensure that the identifier is unique
    login_q = get_login_query(USERS_TABLE, IDENTIFIER_FIELD, username)
    other_users = dal.api_safe_query(login_q)

    if other_users:
        logger.debug(f"The identifier {username} already exists")
        raise HTTPError(400, f"There already exists another user with that {IDENTIFIER_FIELD}")

    # Create the user object, replacing the password with the hashed one
    user = dict(form)
    user[PASSWORD_FIELD] = generate_password_hash(password)

    # Try to insert it in the DB
    # Since the /register endpoint must adapt to any possible table,
    # we assume that the user knows what they're doing and submits the
    # appropriate fields. Otherwise, the DB will just complain.
    register_q = get_register_user_query(USERS_TABLE, user)
    dal.api_safe_update(register_q)

    # Fetch the newly created user from the DB (some fields may have been)
    # automatically generated, like its ID
    # It should now exist
    user = dal.api_safe_query(login_q)[0]

    # If we've reached here the register is successful, generate a session token
    # and return it with the logged user's info
    logger.debug("Register OK")
    token = create_token(user)
    del user[PASSWORD_FIELD]
    res = {"sessionToken": token, "user": user}
    return jsonify(res), 200



import importlib

from os import listdir
from os.path import splitext

from silence.server import manager, default_endpoints
from silence.settings import settings
from silence.logging.default_logger import logger

###############################################################################
# Look for .py files inside the project's "/api" folder
# and force them to run, activating the @endpoint decorators
###############################################################################

def load_user_endpoints():
    logger.debug("Looking for custom endpoints...")

    # Load every .py file inside the api/ folder
    for pyfile in listdir("api"):
        module_name = "api." + splitext(pyfile)[0]
        logger.debug(f"Found endpoint file: {module_name}")

        try:
            importlib.import_module(module_name)
        except ImportError:
            raise RuntimeError(f"Could not load the API file {module_name}")
    
###############################################################################
# Register the Silence-provided endpoints
###############################################################################
def load_default_endpoints():
    route_prefix = settings.API_PREFIX
    if route_prefix.endswith("/"):
        route_prefix = route_prefix[:-1]

    if settings.ENABLE_LOGIN:
        manager.API_TREE.add_url(f"{route_prefix}/login")
        manager.APP.add_url_rule(f"{route_prefix}/login", "login", default_endpoints.login, methods=["POST"])

    if settings.ENABLE_REGISTER:
        manager.API_TREE.add_url(f"{route_prefix}/register")
        manager.APP.add_url_rule(f"{route_prefix}/register", "register", default_endpoints.register, methods=["POST"])



import os
import re

from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

# Borrowed from urllib3 (which is in turn borrowed from SQLAlchemy)
base_path = os.path.dirname(__file__)
ver_path = os.path.join(base_path, "silence", "__init__.py")

with open(ver_path, "r") as f:
    version = re.compile(r""".*__version__ = ["'](.*?)['"]""", re.S).match(f.read()).group(1)

setup(
    name='Silence',
    version=version,
    description='An educational API+Web framework.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/agu-borrego/Silence',
    author='Agustín Borrego',
    author_email='borrego@us.es',
    license='MIT',
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "silence = silence.cli.manager:run_from_command_line",
        ],
    },
    install_requires=[
        'dload==0.6',
        'Flask==1.1.2',
        'Flask-Cors==3.0.8',
        'PyMySQL==0.9.3',
        'Werkzeug==1.0.1',
        'itsdangerous==1.1.0',
        'pypika==0.37.7',
        'colorama==0.4.3',
    ],
    python_requires='~=3.6',
    zip_safe=False
)